


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>

#include "application.h"
#include "backgnd.h"
#include "dlgbox.h"
#include "fft.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "secondwnd.h"



typedef struct
{
	double sum;
	int cnt;
} 
expanded_average;


typedef struct
{
	long red;
	long green;
	long blue;
	int cnt;
} 
expanded_average_rgb;





/********************************************************************

	Eliminate background from main image by moving-average method
	(square of side length 2*radius)
	Colormask is only used for type B_RGB, and it is used to
	mask the panes to be processed. E.g. if rgb.red is nonzero,
	red pane will be processed.
	
	The oorflg determines wheter zero-valued pixels will be
	counted as oor (i.e. background). If oorflg==0, all
	pixels will be processed, otherwise only non-zero
	pixels will be counted for the average computation.
	In the rgb-case, non-background means that at least one
	of r,g,b is non-zero.
	
	THESE ARE 2D OPERATIONS !
	
********************************************************************/	


/* This one computes sum and count in any non-RGB image */


expanded_average sum_area (image_type* img, int x1, int y1, int x2, int y2, int slice, int oorflg)
{
int x,y;
expanded_average a;
float buf;

	a.cnt = 0; a.sum = 0;
	for (x=x1; x<=x2; x++)		/* This order because y dominates */
		for (y=y1; y<=y2; y++)
			if ((x>=0) && (x<img->xmax) && (y>=0) && (y<img->ymax))
			{
				buf = readbuf_flt (img,x,y,slice);
				if ((buf!=0) || (oorflg==0))
				{
					a.sum += buf; a.cnt++;
				}
			}
	return a;

}

/* And this one is for RGB images */

expanded_average_rgb sum_area_rgb (image_type* img, int x1, int y1, int x2, int y2, int slice, int oorflg)
{
int x,y;
expanded_average_rgb a;
rgbtriplet rgb;
rgbtriplet* p;

	a.red = 0; a.green=0; a.blue=0; a.cnt = 0;
	p = img->data;
	for (x=x1; x<=x2; x++)		/* This order because y dominates */
		for (y=y1; y<=y2; y++)
			if ((x>=0) && (x<img->xmax) && (y>=0) && (y<img->ymax))
			{
				rgb = p[ x + img->xmax*(y + img->zmax*slice) ];
				if ((oorflg==0) || (a.red!=0) || (a.green!=0) || (a.blue!=0))
				{
					a.red += rgb.red; a.green += rgb.green; a.blue += rgb.blue; 
					a.cnt++;
				}
			}
	return a;

}



void eliminate_background (image_type* img, int radius, int slice, rgbtriplet colormask, int oorflg)
{
int x,y,x1,x2,y1,y2;
long idx,idx2;
expanded_average a1,a2;
expanded_average_rgb a3,a4;
image_type img2;
double r;
	
	
	/* Clone the current image to hold the high-pass version */
	
	if (allocate_image (&img2, img->imgtype, img->xmax, img->ymax, 1) == -1)
		fatal_memerr();	
	
	/* We try to go strip by strip to accelerate the summation */
	
	for (y=0; y<img->ymax; y++)
	{
		/* Create the initial moving rectangle, once per line.
			To get one IF out of the inner loop, we differentiate here between
			RGB and non-RGB images */
		
		y1 = y-radius; y2=y+radius;
		if (img->imgtype == B_RGB)
		{

			rgbtriplet rgb;
			rgbtriplet* p;
			rgbtriplet* p1;
			p = img->data;
			p1= img2.data;

			a3 = sum_area_rgb (img, -radius, y1, radius, y2, slice, oorflg);
			a4 = a3;
			
			for (x=0; x<img->xmax; x++)
			{
				idx = x + img->xmax*(y + img->ymax*slice);
				idx2= x + img->xmax*y;
				x1 = x-radius; x2=x+radius;
				if (a4.cnt>0)  /* data from the previous pass */
				{
					rgb.red = (int) (a4.red/a4.cnt);
					rgb.green = (int) (a4.green/a4.cnt);
					rgb.blue = (int) (a4.blue/a4.cnt);
				}
				else
				{
					rgb.red=rgb.green=rgb.blue=0;
				}
				
				p1[idx2] = p[idx];
				if (colormask.red)		/* Want to do the red pane. Enter 0 if backgnd woudl cause negative # */
				{
					if (p[idx].red>rgb.red) p1[idx2].red=p[idx].red-rgb.red; else p1[idx2].red=0;
				}
				if (colormask.green)	/* Want to do the green pane. */
				{
					if (p[idx].green>rgb.green) p1[idx2].green=p[idx].green-rgb.green; else p1[idx2].green=0;
				}
				if (colormask.blue)	/* Want to do the green pane. */
				{
					if (p[idx].blue>rgb.blue) p1[idx2].blue=p[idx].blue-rgb.blue; else p1[idx2].blue=0;
				}

				/* Now move the rectangle one pixel to the right */
			
				a3 = sum_area_rgb (img, x2+1,y1,x2+1,y2,slice, oorflg);
				a4.red+=a3.red; a4.green+=a3.green; a4.blue+=a3.blue; a4.cnt+=a3.cnt;
				a3 = sum_area_rgb (img,x1,y1,x1,y2,slice, oorflg);
				a4.red-=a3.red; a4.green-=a3.green; a4.blue-=a3.blue; a4.cnt-=a3.cnt;
			}
		}
		else		/* All non-RGB images */
		{
			a1 = sum_area (img, -radius, y1, radius, y2, slice, oorflg);
			a2 = a1;
			
			for (x=0; x<img->xmax; x++)
			{
				idx = x + img->xmax*(y + img->ymax*slice);
				idx2= x + img->xmax*y;
				x1 = x-radius; x2=x+radius;
				if (a2.cnt>0)  /* data from the previous pass */
					r = a2.sum / a2.cnt;
				else
					r=0;
				
				
				switch (img->imgtype)
				{
					case B_CHAR:
					{
						unsigned char *p,*p1; unsigned char c;
						p=img->data; p1 = img2.data; c = (int)(0.5+r);
						if (p[idx]>c) p1[idx2]=p[idx]-c; else p1[idx2]=0;
						break;
					}
					case B_SHORT:
					{
						short *p, *p1; short c;
						p=img->data; p1 = img2.data; c = (int)(0.5+r);
						if (p[idx]>c) p1[idx2]=p[idx]-c; else p1[idx2]=0;
						break;
					}
					case B_FLOAT:
					{
						float *p, *p1; 
						p=img->data; p1 = img2.data;
						if (p[idx]>r) p1[idx2]=p[idx]-r; else p1[idx2]=0;
						break;
					}
					default:	/* Avoid a warning that B_RGB is not handled here */
					{
					}
				}

				/* Now move the rectangle one pixel to the right */
			
				a1 = sum_area (img, x2+1,y1,x2+1,y2,slice, oorflg);
				a2.sum+=a1.sum; a2.cnt+=a1.cnt;
				a1 = sum_area (img,x1,y1,x1,y2,slice, oorflg);
				a2.sum-=a1.sum; a2.cnt-=a1.cnt;
			}
		
		}
	}
	
	/* Copy the aux image back into the buffer, then discard the aux image */
	
	copybuf (&img2, img, 0,0,0, 0,0,slice, img->xmax * img->ymax);
	freebuf (&img2);
	make_minmax (img);
	img->uid = next_uid++;
}
		


/********************************************************************

	Eliminate background from  image by unsharp masking:
	
	1) Use a strong Gaussian FFT lowpass on a copy of the image
	2) Subtract the lowpass-filtered image from image

	Note that this method cannot handle OOR pixels!
	
	THESE ARE 2D OPERATIONS !
	
********************************************************************/	


void unsharp_masking (image_type* img, int radius, int slice, rgbtriplet colormask, int oorflg)
{
int x,y,e,c;
long idx;
float r;
image_type img2;
float *p;	
rgbtriplet rgb;
int macrorun_save;

	e=1; if (img->imgtype==B_RGB) e=3;
	macrorun_save = macrorun; macrorun=1;			/* Pretend we are calling from a macro */

	/* Clone the current image slice to hold the lowpass (unsharp) version,
		but make it float for better arithmetic later */
		
	if (allocate_image (&img2, B_FLOAT, img->xmax, img->ymax, e) == -1)
		fatal_memerr();	
	p = img2.data;

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if (img->imgtype==B_RGB)			/* Don't apply colormask here, this is minimum additional effort */
			{
				rgb = readbuf_rgb (img, x, y, slice);
				p[x + img->xmax*y] = rgb.red;
				p[x + img->xmax*y + img->xmax*img->ymax] = rgb.green;
				p[x + img->xmax*y + 2*img->xmax*img->ymax] = rgb.blue;
			}
			else
				p[x + img->xmax*y] = readbuf_flt (img,x,y,slice);
		}

	/* Now perform lowpass filtering on the slice (or 3 slices for RGB) */

	fftfilt_go (&img2, 2 /*Gaussian */, 400.0/radius, 1.0, 1.0, 0.0);
	p = img2.data;

	/* Subtract unsharp image from old image INTO img2 */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if (img->imgtype==B_RGB)
			{
				rgb = readbuf_rgb (img, x, y, slice);
				p[x + img->xmax*y] = rgb.red-p[x + img->xmax*y];
				p[x + img->xmax*y + img->xmax*img->ymax] = rgb.green-p[x + img->xmax*y + img->xmax*img->ymax];
				p[x + img->xmax*y + 2*img->xmax*img->ymax] = rgb.blue-p[x + img->xmax*y + 2*img->xmax*img->ymax];
			}
			else
				p[x + img->xmax*y] = readbuf_flt (img,x,y,slice)-p[x + img->xmax*y];
		}

	/* Finally, store filtered image back into img, clamping values if needed. */

	idx = slice*img->xmax*img->ymax;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
				r = readbuf_flt (&img2,x,y,0);
				switch (img->imgtype)
				{
					case B_CHAR:
					{
						unsigned char *p1; 
						p1=img->data; c = (int)(0.5+r);
						if (c>255) c=255; else if (c<0) c=0;
						p1[x + img->xmax*y + idx] = c;
						break;
					}
					case B_SHORT:
					{
						short *p1;
						p1=img->data; c = (int)(0.5+r);
						p1[x + img->xmax*y + idx] = c;
						break;
					}
					case B_FLOAT:
					{
						float *p1; p1=img->data;
						p1[x + img->xmax*y + idx] = r;
						break;
					}
					case B_RGB:
					{
						rgbtriplet *p1; p1 = img->data;
						rgb = p1[x + img->xmax*y + idx];
						c = (int)(0.5+r); if (c>255) c=255; else if (c<0) c=0;
						if (colormask.red) rgb.red=c;
						r = readbuf_flt (&img2,x,y,1);
						c = (int)(0.5+r); if (c>255) c=255; else if (c<0) c=0;
						if (colormask.green) rgb.green=c;
						r = readbuf_flt (&img2,x,y,2);
						c = (int)(0.5+r); if (c>255) c=255; else if (c<0) c=0;
						if (colormask.blue) rgb.blue=c;
						p1[x + img->xmax*y + idx] = rgb;
					}
				}
		}

	freebuf (&img2);
	make_minmax (img);
	img->uid = next_uid++;
	macrorun = macrorun_save;


}



/********************************************************************

	Eliminate background from main image by fitting a plane
	or parabolic shape:
	
	1) Divide the image into squares
	2) Search for minima within the squares, assume 
	   these are background pixels
	3) Fit a background function using least-squares,
	   either B(x,y) = a0 + a1*x + a2*y
	   or     B(x,y) = a0 + a1*x + a2*y + a3*x*x + a4*y*y + a5*x*y
	4) Subtract this background function from the image
	
	THESE ARE 2D OPERATIONS !
	
********************************************************************/	


float bgparms[6];



/* This function fits the 6 parameters into the image slice delivered
   via img. Img has to be B_FLOAT */


int fit_parab_function (image_type* img, int boxsize)
{
int x,y,x1,y1,n,N;
long idx;
double Sx, Sy, Sxx, Syy, Sxy, Sxxx, Syyy, Sxxxx, Syyyy, 
	  Sxxy, Sxyy, Sxxxy, Sxyyy, Sxxyy, Sb, Sbx, Sby, Sbxy, Sbxx, Sbyy;
float min,x2,y2;
float* p;
double mat_a[36]; double b[6];

	g_assert (img->imgtype == B_FLOAT);
	p = img->data;

	N=0; Sx=0.0; Sy=0.0; Sxx=0.0; Syy=0.0; Sxy=0.0; Sxxx=0.0; 
	Syyy=0.0; Sxxxx=0.0; Syyyy=0.0; Sxxy=0.0; Sxyy=0.0; 
	Sxxxy=0.0; Sxyyy=0.0; Sxxyy=0.0; Sb=0.0; Sbx=0.0; Sby=0.0; 
	Sbxx=0.0; Sbyy=0.0; Sbxy=0.0;

	/* Accumulate the intensity/coords for LSQ */

	for (y=0; y<img->ymax; y+=boxsize)
		for (x=0; x<img->xmax; x+=boxsize)
		{
			min = p[x + img->xmax*y]; x2=x/(float)img->xmax; y2=y/(float)img->ymax; 
			for (y1=0; y1<boxsize; y1++)
			{
				for (x1=0; x1<boxsize; x1++)
				{
					if ((x+x1 < img->xmax) && (y+y1 < img->ymax))
					{
						idx = x+x1 + img->xmax*(y+y1);
						if (min>p[idx]) 
						{
							min= p[idx]; 
							x2=((float)x+x1)/img->xmax; y2=((float)y+y1)/img->ymax;
						}
					}
				}
			}
				
			N++; 					Sx+=x2; 				Sy+=y2;
			Sxx += x2*x2; 			Syy += y2*y2; 			Sxy += x2*y2;
			Sxxx += x2*x2*x2; 		Syyy += y2*y2*y2;
			Sxxy += x2*x2*y2; 		Sxyy += x2*y2*y2;
			Sxxxx+=x2*x2*x2*x2; 	Syyyy += y2*y2*y2*y2;
			Sxxxy += x2*x2*x2*y2; 	Sxxyy += x2*x2*y2*y2; 	Sxyyy+= x2*y2*y2*y2;
			Sb += min;
			Sbx += min*x2;
			Sby += min*y2;
			Sbxx+= min*x2*x2;
			Sbyy+= min*y2*y2;
			Sbxy+= min*x2*y2;

		}
		
	dp (1, "n=%d, Sx=%10.3e  Sy=%10.3e  Sxx=%10.3e  Sxy=%10.3e  Syy=%10.3e\n",
		N,Sx,Sy,Sxx,Sxy,Syy);
	dp (1, "Sxxx=%10.3e  Syyy=%10.3e  Sxxy=%10.3e  Sxyy=%10.3e\n",
		Sxxx,Syyy,Sxxy,Sxyy);
	dp (1, "Sxxxx=%10.3e  Syyyy=%10.3e  Sxxxy=%10.3e  Sxxyy=%10.3e  Sxyyy=%10.2e\n",
		Sxxxx,Syyyy,Sxxxy,Sxxyy,Sxyyy);
	
		
	/* Prep the matrix A */
	
	n = 6;	/* 6x6 matrix */
	for (x=0; x<n*n; x++) mat_a[x]=0.0;

	/* Above the diagonal */
	
	mat_a[0]=N; 	mat_a[1]=Sx; 	mat_a[2]=Sy; 	mat_a[3]=Sxx; 	mat_a[4]=Syy; 	mat_a[5]=Sxy;
					mat_a[7]=Sxx; 	mat_a[8]=Sxy; 	mat_a[9]=Sxxx; 	mat_a[10]=Sxyy; mat_a[11]=Sxxy;
									mat_a[14]=Syy; 	mat_a[15]=Sxxy; mat_a[16]=Syyy; mat_a[17]=Sxyy;
													mat_a[21]=Sxxxx;mat_a[22]=Sxxyy;mat_a[23]=Sxxxy;
																	mat_a[28]=Syyyy;mat_a[29]=Sxyyy;
																					mat_a[35]=Sxxyy;
	/*
	A(2,1)=Sx; A(3,1)=Sy; A(4,1)=Sxx; A(5,1)=Syy; A(6,1)=Sxy;
	A(2,2)=Sxx; A(3,2)=Sxy; A(4,2)=Sxxx; A(5,2)=Sxyy; A(6,2)=Sxxy;
	A(3,3)=Syy; A(4,3)=Sxxy; A(5,3)=Syyy; A(6,3)=Sxyy;
	A(4,4)=Sxxxx; A(5,4)=Sxxyy; A(6,4)=Sxxxy;
	A(5,5)=Syyyy; A(6,5)=Sxyyy;
	A(6,6)=Sxxyy;
	*/
	
	/* Mirror to below the diagonal */
	
	for (y=0; y<n; y++)
		for (x=y+1; x<n; x++)
			mat_a[y+n*x] = mat_a[x+n*y];
			
	/* The solution vector */
	
	b[0]=Sb; b[1]=Sbx; b[2]=Sby; b[3]=Sbxx; b[4]=Sbyy; b[5]=Sbxy;

	dp (2, "Here is the matrix equation:\n");
	for (y=0; y<n; y++)
	{
		for (x=0; x<n; x++) dp (2, "%8.3e  ",mat_a[x+n*y]);
		dp (2, "a%1d     %8.3e\n", y, b[y]);
	}
	dp (2, "\n");
	
		
	/* Solve it. */
	
	if (matsolv (mat_a, b, n) < 0)
		return -1;
	
	/* Hand back the result */
	
	for (x=0; x<n; x++)
	{
		bgparms[x]=b[x];
		dp (2, "Parameter a[%d]=%f\n",x,b[x]);
	}
	dp (2, "------------------------------------\n");
	return 0;
	
}
		

int fit_linear_function (image_type* img, int boxsize)
{
int x,y,x1,y1,n,N;
long idx;
float Sx, Sy, Sxx, Syy, Sxy, Sb, Sbx, Sby;
float min,x2,y2;
float* p;
double mat_a[9]; double b[3];

	g_assert (img->imgtype == B_FLOAT);
	p = img->data;

	N=0; Sx=0.0; Sy=0.0; Sxx=0.0; Syy=0.0; Sxy=0.0; 
	Sb=0.0; Sbx=0.0; Sby=0.0;

	/* Accumulate the intensity/coords for LSQ */

	for (y=0; y<img->ymax; y+=boxsize)
		for (x=0; x<img->xmax; x+=boxsize)
		{
			min = p[x + img->xmax*y]; x2=x; y2=y; 
			for (y1=0; y1<boxsize; y1++)
			{
				for (x1=0; x1<boxsize; x1++)
				{
					if ((x+x1 < img->xmax) && (y+y1 < img->ymax))
					{
						idx = x+x1 + img->xmax*(y+y1);
						if (min>p[idx]) 
						{
							min= p[idx];
							x2=((float)x+x1)/img->xmax; y2=((float)y+y1)/img->ymax;
						}
					}
				}
			}
				
			N++; Sx+=x2; Sy+=y2;
			Sxx += SQR(x2); Syy += SQR(y2); Sxy += x2*y2;
			Sb += min; Sbx += min*x2; Sby += min*y2;

		}
	
	/* Prep the matrix A - Note the index shift by 1 through the macro */
	
	n = 3;	/* 3x3 matrix */
	mat_a[0]=N;
	mat_a[1] = Sx; mat_a[3] = Sx;	/* Elements 1,2 and 2,1 */
	mat_a[2] = Sy; mat_a[6] = Sy;	/* Elements 1,3 and 3,1 */
	mat_a[4] = Sxx;					/* Element 2,2 */
	mat_a[5] = Sxy; mat_a[7] = Sxy;	/* Elements 3,2 and 2,3 */
	mat_a[8] = Syy;					/* Element 3,3 */
			
	/* The solution vector */
	
	b[0]=Sb; b[1]=Sbx; b[2]=Sby; 
	
	dp (2, "Here is the matrix equation:\n");
	for (x=0; x<n; x++)
		dp (2, "%8.3e  %8.3e  %8.3e     a%1d     %8.3e\n",
				mat_a[n*x], mat_a[1+x*n], mat_a[2+x*n], x, b[x]);
	dp (2, "\n");
	
	
	/* Solve it. */
	
	if (matsolv (mat_a, b, n) < 0)
		return -1;
	
	/* Hand back the result */
	
	for (x=0; x<6; x++) bgparms[x]=0.0;
	
	for (x=0; x<n; x++)
	{
		bgparms[x]=b[x];
		dp (2, "Parameter a[%d]=%f\n",x,b[x]);
	}
	dp (2, "------------------------------------\n");
	return 0;
	
}
		



void background_fit (image_type* img, int radius, int slice, rgbtriplet colormask, int oorflg, int order)
{
int x,y,j,c;
image_type bk;
float* p;
long idx;
char mask[3];
float x1,y1;

	if (allocate_image (&bk, B_FLOAT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p = bk.data;


	if (img->imgtype==B_RGB)
	{
		unsigned char* p1 = img->data;
		float buf; int i;
		
		if (colormask.red) mask[0]=1; else mask[0]=0;	/* for compatibility */
		if (colormask.green) mask[1]=1; else mask[1]=0;
		if (colormask.blue) mask[2]=1; else mask[2]=0;
		
		idx = slice*img->xmax*img->ymax;
		for (j=0; j<3; j++) if (mask[j])
		{
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
					p[x+img->xmax*y] = p1[j + 3*(idx+x+img->xmax*y)];
	
			if (order==1)
				c = fit_linear_function (&bk, radius);
			else
				c = fit_parab_function (&bk, radius);

			if (c >=0 )
			{
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						x1=(float)x/(float)img->xmax; y1=(float)y/(float)img->ymax;
						buf = bgparms[0] + bgparms[1]*x1 + bgparms[2]*y1
							 + bgparms[3]*x1*x1 + bgparms[4]*y1*y1 + bgparms[5]*y1*x1;
						i = IROUND((float)(p1[j + 3*(idx+x+img->xmax*y)]) - buf);
						if (i<0) i=0; else if (i>255) i=255;
						p1[j + 3*(idx+x+img->xmax*y)] = i;
					}
			}
			
		}
	}
	else
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
				p[x+img->xmax*y] = readbuf_flt (img, x,y,slice);
	
			if (order==1)
				c = fit_linear_function (&bk, radius);
			else
				c = fit_parab_function (&bk, radius);

		if (c >= 0)
		{
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1=(float)x/(float)img->xmax; y1=(float)y/(float)img->ymax;
					p[x+img->xmax*y] = bgparms[0] + bgparms[1]*x1 + bgparms[2]*y1
									 + bgparms[3]*x1*x1 + bgparms[4]*y1*y1 + bgparms[5]*y1*x1;
				}
		}
		else
		{
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
					p[x+img->xmax*y] = 0.0;
		}
		
		make_minmax (&bk);
		if (!macrorun) secondary_display (&bk, 0, "Background");

		switch (img->imgtype)
		{
			case B_CHAR:
			{
				unsigned char* p1 = img->data;
				int i; 
				idx=slice*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						i = IROUND((float)(p1[idx]) - p[x+img->xmax*y]);
						if (i<0) i=0; else if (i>255) i=255;
						p1[idx]=i;
						idx++;
					}
				break;
			}
			case B_SHORT:
			{
				short* p1 = img->data;
				int i; 
				idx=slice*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						i = IROUND((float)(p1[idx]) - p[x+img->xmax*y]);
						p1[idx]=i;
						idx++;
					}
				break;
			}
			case B_FLOAT:
			{
				float* p1 = img->data;
				idx=slice*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						p1[idx] = p1[idx] - p[x+img->xmax*y];
						idx++;
					}
				break;
			}
			
		}	
	
	}
	make_minmax (img);
	img->uid = next_uid++;
	freebuf (&bk);
}


/* Interface for macro interpreter */

void mac_handle_background (image_type* img, int type, int radius, int slice, rgbtriplet colormask, int oorflg)
{

	switch (type)
	{
		case 1:
			eliminate_background (img, radius, slice, colormask, oorflg);
			break;
		case 2:
			unsharp_masking (img, radius, slice, colormask, oorflg);
			break;
		case 3:
			background_fit (img, radius, slice, colormask, oorflg, 1);
			break;
		case 4:
			background_fit (img, radius, slice, colormask, oorflg, 2);
			break;
		default:
		{
		}
	}
}


/*================================================

   U S E R   I N T E R F A C E
   
================================================*/   

GtkWidget *backgnd_window;

GtkAdjustment* bk_adj;
int bk_do_red, bk_do_green, bk_do_blue;
int bk_oorflg, bk_3dflg;
int bkgnd_function;


void on_redbutton_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	bk_do_red = !bk_do_red;
}


void on_greenbutton_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	bk_do_green = !bk_do_green;
}

void on_bluebutton_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	bk_do_blue = !bk_do_blue;
}

void on_oorbutton_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	bk_oorflg = !bk_oorflg;
}

void on_3dbutton_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	bk_3dflg = !bk_3dflg;
}


void on_okbutton_clicked (GtkButton *button, gpointer user_data)
{
rgbtriplet rgb;
int z;

	undo_nexus (&mainimg);

	rgb = (rgbtriplet){0,0,0};
	if (bk_do_red) rgb.red=1;
	if (bk_do_green) rgb.green=1;
	if (bk_do_blue) rgb.blue=1;

	if (bkgnd_function==0)
	{
		if ((mainimg.zmax<2) || (!bk_3dflg))
			eliminate_background (&mainimg, bk_adj->value, cur_slice, rgb, bk_oorflg);
		else
			for (z=0; z<mainimg.zmax; z++)
				eliminate_background (&mainimg, bk_adj->value, z, rgb, bk_oorflg);
	}
	if (bkgnd_function==1)
	{
		if ((mainimg.zmax<2) || (!bk_3dflg))
			unsharp_masking (&mainimg, bk_adj->value, cur_slice, rgb, bk_oorflg);
		else
			for (z=0; z<mainimg.zmax; z++)
				eliminate_background (&mainimg, bk_adj->value, z, rgb, bk_oorflg);
	}
	else if (bkgnd_function==2)
	{
		if ((mainimg.zmax<2) || (!bk_3dflg))
			background_fit (&mainimg, bk_adj->value, cur_slice, rgb, bk_oorflg,1);
		else
			for (z=0; z<mainimg.zmax; z++)
				background_fit (&mainimg, bk_adj->value, z, rgb, bk_oorflg,1);
	}
	else if (bkgnd_function==3)
	{
		if ((mainimg.zmax<2) || (!bk_3dflg))
			background_fit (&mainimg, bk_adj->value, cur_slice, rgb, bk_oorflg,2);
		else
			for (z=0; z<mainimg.zmax; z++)
				background_fit (&mainimg, bk_adj->value, z, rgb, bk_oorflg,2);
	}

	make_imlib_buffer_from_main_image (cur_slice);
	redraw (1);

	gtk_widget_destroy ( GTK_WIDGET ( user_data ) );

}


void on_cancelbutton_clicked (GtkButton *button, gpointer user_data)
{
	gtk_widget_destroy ( GTK_WIDGET ( user_data ) );

}


void bkgnd_set_menu (GtkWidget *widget, gpointer user_data)
{

	bkgnd_function = *(int*) user_data;

}



void create_backgnd_window (void)
{
  GtkWidget *vbox1;
  GtkWidget *redbutton;
  GtkWidget *greenbutton;
  GtkWidget *bluebutton;
  GtkWidget *oorbutton;
  GtkWidget *do3dbutton;
  GtkWidget *hseparator1;
  GtkWidget *hseparator2;
  GtkWidget *label1;
  GtkWidget *hscale1;
  GtkWidget *hbox1;
  GtkWidget *okbutton;
  GtkWidget *cancelbutton;
  GtkWidget *bkgnd_optionmenu1;
  GtkWidget *bkgnd_optionmenu1_menu;
  GtkWidget *glade_menuitem;


	backgnd_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (backgnd_window), "backgnd_window", backgnd_window);
	gtk_window_set_title (GTK_WINDOW (backgnd_window), "Remove Background");
	gtk_window_set_default_size (GTK_WINDOW (backgnd_window), 165, 80);
	/*gtk_widget_set_usize (backgnd_window, 165,80);*/
	gtk_window_set_modal ( GTK_WINDOW ( backgnd_window ), TRUE );
	gtk_container_set_border_width (GTK_CONTAINER (backgnd_window), 3);

	vbox1 = vbox_in_window (backgnd_window,"vbox1");


	bkgnd_optionmenu1 = Make_Optionmenu (backgnd_window, vbox1,"bakgnd_menu",-1,-1);
	gtk_container_set_border_width (GTK_CONTAINER (bkgnd_optionmenu1), 1);
	bkgnd_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (bkgnd_optionmenu1_menu, GTK_SIGNAL_FUNC(bkgnd_set_menu), "Moving Window", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (bkgnd_optionmenu1_menu, GTK_SIGNAL_FUNC(bkgnd_set_menu), "Unsharp masking", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (bkgnd_optionmenu1_menu, GTK_SIGNAL_FUNC(bkgnd_set_menu), "Linear Plane", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (bkgnd_optionmenu1_menu, GTK_SIGNAL_FUNC(bkgnd_set_menu), "Parabolic Plane", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (bkgnd_optionmenu1), bkgnd_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (bkgnd_optionmenu1), 0);
	bkgnd_function=0;
	

	/* The rgb buttons are needed only if the image type is B_RGB */

	if (mainimg.imgtype == B_RGB)
	{
		redbutton = Checkbutton_in_table (backgnd_window,vbox1,"Apply to Red","redbutton",-1,-1,GTK_SIGNAL_FUNC (on_redbutton_toggled));
		greenbutton = Checkbutton_in_table (backgnd_window,vbox1,"Apply to Green","greenbutton",-1,-1,GTK_SIGNAL_FUNC (on_greenbutton_toggled));
		bluebutton = Checkbutton_in_table (backgnd_window,vbox1,"Apply to Blue","bluebutton",-1,-1,GTK_SIGNAL_FUNC (on_bluebutton_toggled));
		hseparator1 = MakeHsep (backgnd_window,vbox1,"hseparator1",-1,-1);
	}
	
	label1 = MakeLabel (backgnd_window,vbox1,"Radius","label1",-1,-1);
	gtk_label_set_justify (GTK_LABEL (label1), GTK_JUSTIFY_LEFT);

	bk_adj = GTK_ADJUSTMENT (gtk_adjustment_new (40, 0, 200, 1, 10, 10));
	hscale1 = MakeHScale (GTK_ADJUSTMENT (bk_adj),backgnd_window,vbox1,"hscale1",-1,-1);

	oorbutton = Checkbutton_in_table (backgnd_window,vbox1,"Zero is \"out of region\"","oorbutton",-1,-1,GTK_SIGNAL_FUNC (on_oorbutton_toggled));

	if (mainimg.zmax>0)
		do3dbutton = Checkbutton_in_table (backgnd_window,vbox1,"Perform for all slices","do3drbutton",-1,-1,GTK_SIGNAL_FUNC (on_3dbutton_toggled));

	hseparator2 = MakeHsep (backgnd_window,vbox1,"hseparator1",-1,-1);

	hbox1 = gtk_hbox_new (TRUE, 0);
	gtk_widget_ref (hbox1);
	gtk_object_set_data_full (GTK_OBJECT (backgnd_window), "hbox1", hbox1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox1);
	gtk_box_pack_start (GTK_BOX (vbox1), hbox1, TRUE, TRUE, 7);

	okbutton = Button_in_table (backgnd_window,hbox1,"OK","okbutton",-1,-1, NULL);
	cancelbutton = Button_in_table (backgnd_window,hbox1,"Cancel","cancelbutton",-1,-1, NULL);

	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                    	GTK_SIGNAL_FUNC (on_okbutton_clicked), backgnd_window);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                    	GTK_SIGNAL_FUNC (on_cancelbutton_clicked), backgnd_window);
	gtk_signal_connect (GTK_OBJECT (backgnd_window), "destroy",
                    	GTK_SIGNAL_FUNC (on_cancelbutton_clicked), backgnd_window);

  	gtk_widget_show (backgnd_window);
}


/* Create the UI, start the process */

void background_main ()
{

	bk_do_red = bk_do_green = bk_do_blue = 0;
	bk_oorflg = 0; bk_3dflg = 0;
	create_backgnd_window ();
	
}  


